# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright (C) Albert Kottke, 2013-2015
import numpy as np
import pytest
from numpy.testing import assert_allclose
from numpy.testing import assert_equal
from scipy.integrate import trapz

from . import FPATH_DATA
from pysra import motion


@pytest.fixture
def tsm():
    """Create a default time series for testing."""
    return motion.TimeSeriesMotion.load_at2_file(FPATH_DATA / "NIS090.AT2")


def test_ts_load_at2_file(tsm):
    """Test loading of an AT2 file."""
    assert_equal(tsm.accels.size, 4096)
    assert_allclose(tsm.time_step, 0.01)
    assert_allclose(tsm.accels[0], 0.233833e-06)
    assert_allclose(tsm.accels[-1], 0.496963e-04)


def test_ts_times(tsm):
    """Test times."""
    assert_allclose(
        [tsm.times[0], tsm.times[1], tsm.times[-1]],
        [0, tsm.time_step, tsm.time_step * (len(tsm.accels) - 1)],
    )


def test_ts_freqs(tsm):
    """Test calculation of a time series frequencies."""
    freqs = tsm.freqs
    assert_equal(tsm.freqs.size, tsm.fourier_amps.size)
    assert_allclose(freqs[0], 0)
    assert_allclose(freqs[-1], 50.0)


def test_ts_max(tsm):
    """Check that maximum is consistent with the time domain maximum (PGA)."""
    assert_allclose(tsm.pga, tsm.calc_peak())


def test_ts_fft(tsm):
    """Test FFT with no transfer function."""
    assert_allclose(
        tsm.accels,
        tsm.calc_time_series(),
    )


def test_ts_fft_with_tf(tsm):
    """Test FFT with a transfer function."""
    assert_allclose(
        tsm.accels,
        tsm.calc_time_series(2 * np.ones_like(tsm.freqs)) / 2,
    )


def test_ts_fa_normalize(tsm):
    """Test the normalization of the Fourier amplitudes with Parsevla's theorem."""
    assert_allclose(
        trapz(tsm.accels ** 2, dx=tsm.time_step),
        2 * trapz(np.abs(tsm.fourier_amps) ** 2, tsm.freqs),
    )


@pytest.mark.parametrize("fname", ["2516b_a.smc"])
def test_ts_load_smc_file(fname):
    tsm = motion.TimeSeriesMotion.load_smc_file(FPATH_DATA / fname)
    assert tsm.description == "VA: Reston; Fire Station #25; 360"

    assert_allclose(tsm.time_step, 1 / 200.0)

    assert_allclose(
        [tsm.accels[0], tsm.accels[1], tsm.accels[-1]],
        [2.3489e-2, -1.6646e-2, 3.4990e-3],
        rtol=1e-4,
    )



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright (C) Albert Kottke, 2013-2022
import json
import string

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pyexcel
import pytest

import pysra
from . import FPATH_DATA


def read_cluster(ws, cols, names, row_start, row_end):
    d = dict()
    for c, name in zip(cols, names):
        range_str = "{col}{start}:{col}{end}".format(
            col=c, start=row_start, end=row_end
        )
        d[name] = [row[0].value for row in ws[range_str]]
    return d


def read_deepsoil_results(name):
    data = pyexcel.get_array(file_name=str(FPATH_DATA / (name + ".xlsx")))
    names = ",".join(string.ascii_uppercase[: len(data[0])])
    records = np.rec.fromrecords(data, names=names)

    def extract_cols(records, cols, first, last, names):
        return {
            name: records[col][first:last].astype(float)
            for col, name in zip(cols, names)
        }

    d = dict()
    # Read the time series
    d["time_series"] = extract_cols(
        records, "ABCDE", 1, 11800, ["time", "accel", "strain", "stress", "arias_int"]
    )
    # Read the response spectrum
    d["resp_spec"] = extract_cols(records, "GH", 1, 114, ["period", "psa"])
    # Read the Fourier amplitude
    d["fourier_spec"] = extract_cols(
        records, "JKL", 1, 16384, ["freq", "ampl", "ratio"]
    )

    return d


def load_ts():
    fpath = FPATH_DATA / "ChiChi.txt"
    with fpath.open() as fp:
        parts = next(fp).split()
        time_step = float(parts[1])
        accels = [float(line.split()[1]) for line in fp]

    return pysra.motion.TimeSeriesMotion(
        fpath.name, "ChiChi.txt from DeepSoil v6.1", time_step, accels
    )


class DeepSoilComparison:
    rtol = 0.005
    atol = 0.001

    ref_name = NotImplemented
    calc = NotImplemented
    profile = NotImplemented

    @classmethod
    def setup_class(cls):
        cls.ref = read_deepsoil_results(cls.ref_name)
        # Perform the calculation
        cls.calc(load_ts(), cls.profile, cls.profile.location("outcrop", index=-1))
        cls.outputs = pysra.output.OutputCollection(
            [
                pysra.output.AccelerationTSOutput(
                    pysra.output.OutputLocation("outcrop", depth=0)
                ),
                pysra.output.AriasIntensityTSOutput(
                    pysra.output.OutputLocation("outcrop", depth=0)
                ),
                pysra.output.StrainTSOutput(
                    pysra.output.OutputLocation("within", depth=10), in_percent=True
                ),
                pysra.output.StressTSOutput(
                    pysra.output.OutputLocation("within", depth=10), normalized=True
                ),
                pysra.output.ResponseSpectrumOutput(
                    [
                        100.0000,
                        93.9744,
                        88.3119,
                        82.9910,
                        77.9903,
                        73.2912,
                        68.8753,
                        64.7249,
                        60.8250,
                        57.1602,
                        53.7158,
                        50.4793,
                        47.4377,
                        44.5794,
                        41.8932,
                        39.3690,
                        36.9969,
                        34.7676,
                        32.6727,
                        30.7040,
                        28.8540,
                        27.1154,
                        25.4816,
                        23.9462,
                        22.5034,
                        21.1474,
                        19.8732,
                        18.6757,
                        17.5504,
                        16.4930,
                        15.4992,
                        14.5653,
                        13.6877,
                        12.8629,
                        12.0879,
                        11.3595,
                        10.6751,
                        10.0319,
                        9.4274,
                        8.8594,
                        8.3256,
                        7.8239,
                        7.3525,
                        6.9094,
                        6.4931,
                        6.1019,
                        5.7342,
                        5.3887,
                        5.0640,
                        4.7589,
                        4.4721,
                        4.2027,
                        3.9494,
                        3.7115,
                        3.4878,
                        3.2777,
                        3.0802,
                        2.8946,
                        2.7202,
                        2.5563,
                        2.4022,
                        2.2575,
                        2.1215,
                        1.9936,
                        1.8735,
                        1.7606,
                        1.6545,
                        1.5549,
                        1.4612,
                        1.3731,
                        1.2904,
                        1.2126,
                        1.1396,
                        1.0709,
                        1.0064,
                        0.9457,
                        0.8888,
                        0.8352,
                        0.7849,
                        0.7376,
                        0.6931,
                        0.6514,
                        0.6121,
                        0.5752,
                        0.5406,
                        0.5080,
                        0.4774,
                        0.4486,
                        0.4216,
                        0.3962,
                        0.3723,
                        0.3499,
                        0.3288,
                        0.3090,
                        0.2904,
                        0.2729,
                        0.2564,
                        0.2410,
                        0.2265,
                        0.2128,
                        0.2000,
                        0.1879,
                        0.1766,
                        0.1660,
                        0.1560,
                        0.1466,
                        0.1378,
                        0.1295,
                        0.1217,
                        0.1143,
                        0.1074,
                        0.1010,
                        0.1000,
                    ],
                    pysra.output.OutputLocation("outcrop", depth=0),
                    0.05,
                ),
            ]
        )
        cls.outputs(cls.calc)

    def test_times(self):
        ref = self.ref["time_series"]["time"]
        n = len(ref)
        np.testing.assert_allclose(
            self.outputs[0].refs[:n], ref, rtol=self.rtol, atol=self.atol
        )

    def test_accels(self):
        ref = self.ref["time_series"]["time"]
        n = len(ref)
        np.testing.assert_allclose(
            self.outputs[0].refs[:n], ref, rtol=self.rtol, atol=self.atol
        )

    def test_arias_ints(self):
        ref = self.ref["time_series"]["arias_int"]
        n = len(ref)
        np.testing.assert_allclose(
            self.outputs[1].values[:n], ref, rtol=self.rtol, atol=self.atol
        )

    def test_strains(self):
        ref = self.ref["time_series"]["strain"]
        n = len(ref)
        np.testing.assert_allclose(
            self.outputs[2].values[:n], ref, rtol=self.rtol, atol=self.atol
        )

    def test_stresses(self):
        ref = self.ref["time_series"]["stress"]
        n = len(ref)
        np.testing.assert_allclose(
            self.outputs[3].values[:n], ref, rtol=self.rtol, atol=self.atol
        )

    def test_periods(self):
        np.testing.assert_allclose(
            self.outputs[4].periods,
            self.ref["resp_spec"]["period"],
            rtol=self.rtol,
            atol=self.atol,
        )

    def test_osc_freqs(self):
        np.testing.assert_allclose(
            1 / self.outputs[4].refs,
            self.ref["resp_spec"]["period"],
            rtol=self.rtol,
            atol=self.atol,
        )

    def test_spec_accels(self):
        np.testing.assert_allclose(
            self.outputs[4].values,
            self.ref["resp_spec"]["psa"],
            rtol=self.rtol,
            atol=self.atol,
        )


class TestExample02LE(DeepSoilComparison):
    # Test the linear elastic wave propagation
    ref_name = "ds-example-2a-le"
    calc = pysra.propagation.LinearElasticCalculator()
    profile = pysra.site.Profile(
        [
            pysra.site.Layer(
                pysra.site.SoilType("Soil", 20.0, mod_reduc=None, damping=0), 20, 500
            ),
            pysra.site.Layer(
                pysra.site.SoilType("Rock", 25.0, mod_reduc=None, damping=0.02), 0, 760
            ),
        ]
    )


class TestExample02EL(DeepSoilComparison):
    # Run the linear elastic test with the EL calculator.
    ref_name = "ds-example-2a-le"
    calc = pysra.propagation.EquivalentLinearCalculator()
    profile = pysra.site.Profile(
        [
            pysra.site.Layer(
                pysra.site.SoilType("Soil", 20.0, mod_reduc=None, damping=0), 20, 500
            ),
            pysra.site.Layer(
                pysra.site.SoilType("Rock", 25.0, mod_reduc=None, damping=0.02), 0, 760
            ),
        ]
    )


class TestExample04EL(DeepSoilComparison):
    ref_name = "ds-example-4-eql"
    calc = pysra.propagation.EquivalentLinearCalculator()
    profile = pysra.site.Profile(
        [
            pysra.site.Layer(
                pysra.site.SoilType("Soil", 20.0, mod_reduc=None, damping=0.05), 20, 500
            ),
            pysra.site.Layer(
                pysra.site.SoilType("Rock", 25.0, mod_reduc=None, damping=0.02), 0, 760
            ),
        ]
    )


class QWLComparison:
    rtol = 0.01
    atol = 0.01

    index = NotImplementedError

    @classmethod
    def setup_class(cls):
        fpath = FPATH_DATA / "qwl_tests.json"
        data = json.load(fpath.open())[cls.index]

        thickness = np.diff(data["site"]["depth"])

        profile = pysra.site.Profile()
        for i, thick in enumerate(thickness):
            if "damping" in data["site"]:
                damping = data["site"]["damping"][i]
            else:
                damping = None

            profile.append(
                pysra.site.Layer(
                    pysra.site.SoilType(
                        f"{i}",
                        data["site"]["density"][i] * pysra.motion.GRAVITY,
                        damping=damping,
                    ),
                    thick * 1000,
                    data["site"]["velocity"][i] * 1000,
                )
            )

        profile.update_layers()

        if "site_atten" in data["site"]:
            site_atten = data["site"]["site_atten"]
        else:
            site_atten = profile.site_attenuation()

        cls.motion = pysra.motion.Motion(data["freqs"])
        cls.calc = pysra.propagation.QuarterWaveLenCalculator(site_atten=site_atten)
        cls.calc(cls.motion, profile, profile.location("outcrop", index=-1))
        cls.data = data

    def test_crustal_amp(self):
        ref = self.data["crustal_amp"]
        np.testing.assert_allclose(
            self.calc.crustal_amp, ref, rtol=self.rtol, atol=self.atol
        )

    def test_site_term(self):
        ref = self.data["site_term"]
        np.testing.assert_allclose(
            self.calc.site_term, ref, rtol=self.rtol, atol=self.atol
        )


class TestQwl0(QWLComparison):
    index = 0


# Not sure why this test fails. I suspect there is a typo in one of the
# equations
@pytest.mark.xfail
class TestQwl1(QWLComparison):
    index = 1


# class TestQwl2(QWLComparison):
#     rtol = 0.05
#     index = 2


def test_quarter_wavelength_fit():
    fpath = FPATH_DATA / "qwl_tests.json"
    data = json.load(fpath.open())[0]
    thickness = np.diff(data["site"]["depth"])

    profile = pysra.site.Profile()
    for i, (thick, vel, density) in enumerate(
        zip(thickness, data["site"]["velocity"], data["site"]["density"])
    ):
        profile.append(
            pysra.site.Layer(
                pysra.site.SoilType(f"{i}", density * pysra.motion.GRAVITY),
                thick * 1000,
                vel * 1000,
            )
        )

    profile.update_layers()

    motion = pysra.motion.Motion(data["freqs"])
    calc = pysra.propagation.QuarterWaveLenCalculator(
        site_atten=data["site"]["site_atten"]
    )
    calc(motion, profile, profile.location("outcrop", index=-1))

    calc.fit("crustal_amp", data["crustal_amp"])

    np.testing.assert_allclose(
        profile.initial_shear_vel, calc.profile.initial_shear_vel, rtol=0.2
    )


def test_linear_elastic_nrattle():
    """Test against nrattle."""
    ctl = pysra.tools.read_nrattle_ctl(FPATH_DATA / "nrattle.ctl")
    results = np.rec.fromrecords(
        np.loadtxt(
            FPATH_DATA / "test_nrattle_02mar12.nrattle_amps4plot.out",
            skiprows=19,
            usecols=(0, 2),
        ),
        names="freq,amp",
    )

    # Create pysra objects
    profile = pysra.tools.profile_from_nrattle_ctl(ctl)
    motion = pysra.motion.Motion(freqs=results.freq)
    calc = pysra.propagation.LinearElasticCalculator()

    loc_in = profile.location("outcrop", index=(ctl["hs_layer"] - 1))
    loc_out = profile.location("outcrop", depth=ctl["out_depth"])

    calc(
        motion,
        profile,
        # Need to convert to zero-based indices
        profile.location("outcrop", index=(ctl["hs_layer"] - 1)),
    )
    trans_func = np.abs(calc.calc_accel_tf(loc_in, loc_out))
    # Data is provided to 4 decimal places
    rounded = np.round(trans_func, 4)

    if 0:
        # Make a plot for debugging
        fig = plt.figure(constrained_layout=True)
        spec = fig.add_gridspec(nrows=3)

        ax = fig.add_subplot(spec[:2])
        ax.plot(results.freq, results.amp, label="nRattle")
        ax.plot(motion.freqs, trans_func, label="pysra", ls="--")

        ax.legend()
        ax.set(xscale="linear", ylabel="|Transfer Func.|")

        ax = fig.add_subplot(spec[2])
        ax.plot(results.freq, (rounded - results.amp) / results.amp)
        ax.set_label("Rel. Diff.")
        ax.set(xlabel="Frequency (Hz)", xscale="linear", ylabel="Rel. Diff.")

        fig.savefig("test_propagation-test_linear_elastic_nrattle.png", dpi=200)

    np.testing.assert_allclose(
        results.amp,
        rounded,
        rtol=1e-4,
    )


#
# def compare_ts_results(calc, name):
#     ref_soil, ref_inp = (name)
#
#     # Compare the time series
#     # Only compare the number of values in the DeepSoil results,
#     # which doesn't include the zero padding added by the FFT.
#     n = len(ref_soil['time_series']['accel'])
#     loc_surface = pysra.output.OutputLocation('outcrop', index=0)
#     loc_midheight = pysra.output.OutputLocation(
#         'within', depth=(calc.profile[0].thickness / 2))
#     for key, output in [
#         ('accel',),
#         ('arias_int',
#          pysra.output.AriasIntensityTSOutput(loc_surface)),
#         ('strain', pysra.output.StrainTSOutput(loc_midheight)),
#         ('stress', pysra.output.StressTSOutput(loc_midheight, damped=False)),
#     ]:
#         output(calc)
#         import matplotlib.pyplot as plt
#         fig, ax = plt.subplots()
#         ax.plot(
#             ref_soil['time_series']['time'],
#             ref_soil['time_series'][key], 'b-'
#         )
#         ax.plot(output.refs, output.values, 'r--')
#         ax.set_xlim(10, 40)
#         fig.tight_layout()
#         fig.savefig('test')
#
#
#
# def test_linear(ts):
#     calc =
#     calc(ts, profile, profile.location('outcrop', index=-1))
#     compare_ts_results(calc, '')
#
#
# def test_equiv_linear():
#     pass



# The MIT License (MIT)
#
# Copyright (c) 2016-2018 Albert Kottke
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import copy
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
from scipy import stats
from scipy.sparse import diags

from . import site

# Used to define the random state. A specific state can be set with:
# random_state.set_seed(42)
random_state = np.random.RandomState()


class TruncatedNorm:
    """Truncated normal random number generator.

    Parameters
    ----------
    limit : float
        Standard normal limits to impose
    """

    def __init__(self, limit):
        self.limit = limit

    @property
    def limit(self):
        return self._limit

    @limit.setter
    def limit(self, value):
        self._limit = value

        # Need to scale the standard deviation to achieve sample standard
        # deviation based on the truncation. Given truncation of 2 standard
        # deviations, the input standard deviation must be increased to
        # 1.136847 to maintain a unit standard deviation for the random
        # samples.
        self._scale = 1 / np.sqrt(stats.truncnorm.stats(-value, value, moments="v"))

    @property
    def scale(self):
        return self._scale

    def __call__(self, size=1):
        """Random number generator that follows a truncated normal distribution.

        This is the defalut random number generator used by the program. It
        generates normally distributed values ranging from -2 to +2 with unit
        standard deviation.

        The state of the random number generator is controlled by the
        ``np.random.RandomState`` instance.

        Parameters
        ----------
        size : int
            Number of random values to compute

        Returns
        -------
        rvs : ndarray or scalar
            Random variates of given `size`.
        """
        return stats.truncnorm.rvs(
            -self.limit, self.limit, scale=self._scale, size=size
        )

    def correlated(self, correl):
        # Acceptance proportion
        accept = np.diff(stats.norm.cdf([-self.limit, self.limit]))[0]
        # The expected number of tries required
        expected = np.ceil(1 / accept).astype(int)

        while True:
            # Compute the multivariate normal with a unit variance and
            # specified standard deviation. Use twice the expected since
            # this calculation is fast and we don't want to loop.
            randvar = np.random.multivariate_normal(
                [0, 0], [[1, correl], [correl, 1]], size=(2 * expected)
            )
            valid = np.all(np.abs(randvar) < self.limit, axis=1)
            if np.any(valid):
                # Return the first valid value
                return randvar[valid][0]


# Random number generator used for all random number. Limited to +/- 2,
# and the standard deviation is scaled to maintain the standard deviation
# FIXME
randnorm = TruncatedNorm(2)


class ToroThicknessVariation(object):
    """Toro (1995) [T95]_ thickness variation model.

    The recommended values are provided as defaults to this model.

    .. rubric:: References

    .. [T95] Toro, G. R. (1995). Probabilistic models of site velocity
        profiles for generic and site-specific ground-motion amplification
        studies. Brookhaven National Laboratory Technical Report: 779574.

    Parameters
    ----------
    c_1: float, optional
        :math:`c_1` model parameter.
    c_2: float, optional
        :math:`c_2` model parameter.
    c_3: float, optional
        :math:`c_3` model parameter.

    """

    def __init__(self, c_1=10.86, c_2=-0.89, c_3=1.98):
        self._c_1 = c_1
        self._c_2 = c_2
        self._c_3 = c_3

    @property
    def c_3(self):
        return self._c_3

    @property
    def c_2(self):
        return self._c_2

    @property
    def c_1(self):
        return self._c_1

    def iter_thickness(self, depth_total):
        """Iterate over the varied thicknesses.

        The layering is generated using a non-homogenous Poisson process. The
        following routine is used to generate the layering. The rate
        function, :math:`\lambda(t)`, is integrated from 0 to t to generate
        cumulative rate function, :math:`\Lambda(t)`. This function is then
        inverted producing :math:`\Lambda^{-1}(t)`. Random variables
        are produced using the a exponential random variation with
        :math:`\mu = 1` and converted to the nonhomogenous variables using
        the inverted function.

        Parameters
        ----------
        depth_total: float
            Total depth generated. Last thickness is truncated to achieve
            this depth.

        Yields
        ------
        float
            Varied thickness.

        """
        total = 0
        depth_prev = 0

        while depth_prev < depth_total:
            # Add a random exponential increment
            total += np.random.exponential(1.0)

            # Convert between x and depth using the inverse of \Lambda(t)
            depth = (
                np.power(
                    (self.c_2 * total) / self.c_3
                    + total / self.c_3
                    + np.power(self.c_1, self.c_2 + 1),
                    1 / (self.c_2 + 1),
                )
                - self.c_1
            )

            thickness = depth - depth_prev

            if depth > depth_total:
                thickness = depth_total - depth_prev
                depth = depth_prev + thickness

            depth_mid = (depth_prev + depth) / 2
            yield thickness, depth_mid

            depth_prev = depth

    def __call__(self, profile):
        """Calculated a varied thickness profile.

        Parameters
        ----------
        profile : site.Profile
            Profile to be varied. Not modified in place.

        Returns
        -------
        site.Profile
            Varied site profile.

        """
        layers = []
        for (thick, depth_mid) in self.iter_thickness(profile[-2].depth_base):
            # Locate the proper layer and add it to the model
            for l in profile:
                if l.depth < depth_mid <= l.depth_base:
                    layers.append(site.Layer(l.soil_type, thick, l.initial_shear_vel))
                    break
            else:
                raise LookupError

        # Add the half-space
        hsl = profile[-1]
        layers.append(site.Layer(hsl.soil_type, 0, hsl.initial_shear_vel))

        varied = site.Profile(layers, profile.wt_depth)
        return varied


class BedrockDepthVariation(object):
    def __init__(
        self,
        dist: stats.rv_continuous,
        discretize_kwds: Optional[Dict[str, float]] = None,
    ):
        self._dist = dist
        self._discretize_kwds = discretize_kwds or dict(max_freq=50, wave_frac=2)

    def __call__(self, profile: site.Profile) -> site.Profile:
        varied_depth = self._dist.rvs()

        index, depth_within = profile.lookup_depth(varied_depth)

        _profile = site.Profile([layer for layer in profile[: (index + 1)]])
        _profile[-1].thickness = depth_within

        return _profile.auto_discretize(**self._discretize_kwds)


class LayerThicknessVariation(object):
    def __init__(
        self,
        models: Union[List[stats.rv_continuous], Dict[int, stats.rv_continuous]],
        discretize_kwds: Optional[Dict[str, float]] = None,
    ) -> None:

        self._models = models
        self._discretize_kwds = discretize_kwds


class VelocityVariation(object):
    """Abstract model for varying the velocity."""

    def __init__(self, vary_bedrock=False):
        self._vary_bedrock = vary_bedrock

    def __call__(self, profile: site.Profile) -> site.Profile:
        """Calculate a varied shear-wave velocity profile.

        Parameters
        ----------
        profile : site.Profile
            Profile to be varied. Not modified in place.

        Returns
        -------
        site.Profile
            Varied site profile.
        """

        mean = np.log(profile.initial_shear_vel)
        covar = self._calc_covar_matrix(profile)

        ln_vel_rand = np.random.multivariate_normal(mean, covar, check_valid="ignore")

        # Limits based on the number of standard deviations
        offset = randnorm.limit * np.sqrt(np.diag(covar))
        ln_vel = np.clip(ln_vel_rand, mean - offset, mean + offset)
        vel = np.exp(ln_vel)

        # Create the new layers
        end = None if self.vary_bedrock else -1
        # FIXME: Can we use .copy()?
        layers = [
            site.Layer(l.soil_type, l.thickness, vel[i])
            for i, l in enumerate(profile[:end])
        ]
        if not self.vary_bedrock:
            layers.append(profile[-1])

        varied = site.Profile(layers, profile.wt_depth)
        return varied

    def _calc_covar_matrix(self, profile):
        """Calculate the covariance matrix.

        Parameters
        ----------
        profile : site.Profile
            Input site profile

        Yields
        ------
        covar : `class`:numpy.array
            Covariance matrix
        """
        corr = self._calc_corr(profile)
        std = self._calc_ln_std(profile)
        # Modify the standard deviation by the truncated norm scale
        std *= randnorm.scale

        var = std**2
        covar = corr * std[:-1] * std[1:]

        # Main diagonal is the variance
        mat = diags([covar, var, covar], [-1, 0, 1]).toarray()

        return mat

    def _calc_corr(self, profile):
        """Compute the adjacent-layer correlations.

        Parameters
        ----------
        profile : site.Profile
            Input site profile

        Yields
        ------
        np.array
            Correlation matrix
        """
        raise NotImplementedError

    def _calc_ln_std(self, profile):
        """Compute the standard deviation for each layer.

        Parameters
        ----------
        profile : site.Profile
            Input site profile

        Yields
        ------
        np.array
            Standard deviation of the shear-wave velocity
        """
        raise NotImplementedError

    @property
    def vary_bedrock(self):
        return self._vary_bedrock


class ToroVelocityVariation(VelocityVariation):
    """Toro (1995) [T95] velocity variation model.

    Default values can be selected with :meth:`.generic_model`.

    Parameters
    ----------
    ln_std: float, optional
        :math:`\sigma_{ln}` model parameter.
    rho_0: float, optional
        :math:`ρ_0` model parameter.
    delta: float, optional
        :math:`\Delta` model parameter.
    rho_200: float, optional
        :math:`ρ_200` model parameter.
    h_0: float, optional
        :math:`h_0` model parameter.
    b: float, optional
        :math:`b` model parameter.
    """

    PARAMS = {
        "Geomatrix AB": {
            "ln_std": 0.46,
            "rho_0": 0.96,
            "delta": 13.1,
            "rho_200": 0.96,
            "h_0": 0.0,
            "b": 0.095,
        },
        "Geomatrix CD": {
            "ln_std": 0.38,
            "rho_0": 0.99,
            "delta": 8.0,
            "rho_200": 1.00,
            "h_0": 0.0,
            "b": 0.160,
        },
        "USGS AB": {
            "ln_std": 0.35,
            "rho_0": 0.95,
            "delta": 4.2,
            "rho_200": 1.00,
            "h_0": 0.0,
            "b": 0.138,
        },
        "USGS CD": {
            "ln_std": 0.36,
            "rho_0": 0.99,
            "delta": 3.9,
            "rho_200": 1.00,
            "h_0": 0.0,
            "b": 0.293,
        },
        "USGS A": {
            "ln_std": 0.36,
            "rho_0": 0.95,
            "delta": 3.4,
            "rho_200": 0.42,
            "h_0": 0.0,
            "b": 0.063,
        },
        "USGS B": {
            "ln_std": 0.27,
            "rho_0": 0.97,
            "delta": 3.8,
            "rho_200": 1.00,
            "h_0": 0.0,
            "b": 0.293,
        },
        "USGS C": {
            "ln_std": 0.31,
            "rho_0": 0.99,
            "delta": 3.9,
            "rho_200": 0.98,
            "h_0": 0.0,
            "b": 0.344,
        },
        "USGS D": {
            "ln_std": 0.37,
            "rho_0": 0.00,
            "delta": 5.0,
            "rho_200": 0.50,
            "h_0": 0.0,
            "b": 0.744,
        },
    }

    def __init__(
        self,
        ln_std: float,
        rho_0: float,
        delta: float,
        rho_200: float,
        h_0: float,
        b: float,
        vary_bedrock: bool = False,
    ):
        """Initialize the model."""
        super().__init__(vary_bedrock=vary_bedrock)

        self._ln_std = ln_std
        self._rho_0 = rho_0
        self._delta = delta
        self._rho_200 = rho_200
        self._h_0 = h_0
        self._b = b

    def _calc_corr(self, profile: site.Profile) -> np.ndarray:
        """Compute the adjacent-layer correlations

        Parameters
        ----------
        profile : :class:`site.Profile`
            Input site profile

        Yields
        ------
        corr : :class:`numpy.array`
            Adjacent-layer correlations
        """
        depth = profile.depth_mid[:-1]
        thick = np.diff(depth)
        depth = depth[1:]

        # Depth dependent correlation
        corr_depth = self.rho_200 * np.power(
            (depth + self.rho_0) / (200 + self.rho_0), self.b
        )
        corr_depth[depth > 200] = self.rho_200

        # Thickness dependent correlation
        corr_thick = self.rho_0 * np.exp(-thick / self.delta)

        # Final correlation
        # Correlation coefficient
        corr = (1 - corr_depth) * corr_thick + corr_depth

        # Bedrock is perfectly correlated with layer above it
        corr = np.r_[corr, 1]

        return corr

    def _calc_ln_std(self, profile):
        ln_std = self.ln_std * np.ones(len(profile))
        return ln_std

    @property
    def ln_std(self):
        return self._ln_std

    @property
    def rho_0(self):
        return self._rho_0

    @property
    def delta(self):
        return self._delta

    @property
    def rho_200(self):
        return self._rho_200

    @property
    def h_0(self):
        return self._h_0

    @property
    def b(self):
        return self._b

    @classmethod
    def site_classes(cls):
        return cls.PARAMS.keys()

    @classmethod
    def generic_model(cls, site_class, **kwds):
        """Use generic model parameters based on site class.

        Parameters
        ----------
        site_class: str
            Site classification. Possible options are:
             * Geomatrix AB
             * Geomatrix CD
             * USGS AB
             * USGS CD
             * USGS A
             * USGS B
             * USGS C
             * USGS D

            See the report for definitions of the Geomatrix site
            classication. USGS site classification is based on :math:`V_{s30}`:

            =========== =====================
            Site Class  :math:`V_{s30}` (m/s)
            =========== =====================
            A           >750 m/s
            B           360 to 750 m/s
            C           180 to 360 m/s
            D           <180 m/s
            =========== =====================

        Returns
        -------
        ToroVelocityVariation
            Initialized :class:`ToroVelocityVariation` with generic parameters.
        """
        p = dict(cls.PARAMS[site_class])
        p.update(kwds)
        return cls(**p)


class DepthDependToroVelVariation(ToroVelocityVariation):
    """Toro (1995) [T95] velocity variation model modified for a depth
    dependent standard deviation that can be overridden by the soil_type name.

    Default values can be selected with :meth:`.generic_model`.

    Parameters
    ----------
    ln_std_map: dict[str, float], optional
        Mapping between the soil_type and the defined ln_std. Default is *None*.
    depth: array_like, optional
        Depths defining the standard deviation model. Default is [0, 15]
        following the SPID model.
    ln_std: array_like, optional
        :math:`\sigma_{ln}` model parameter. Default is [0.25, 0.15]
        following the SPID model.
    rho_0: float, optional
        :math:`ρ_0` model parameter.
    delta: float, optional
        :math:`\Delta` model parameter.
    rho_200: float, optional
        :math:`ρ_200` model parameter.
    h_0: float, optional
        :math:`h_0` model parameter.
    b: float, optional
        :math:`b` model parameter.
    """

    def __init__(
        self,
        ln_std_map,
        depth,
        ln_std,
        rho_0,
        delta,
        rho_200,
        h_0,
        b,
        vary_bedrock=False,
    ):
        """Initialize the model."""
        super().__init__(
            depth, ln_std, rho_0, delta, rho_200, h_0, b, vary_bedrock=vary_bedrock
        )
        self._ln_std_map = ln_std_map or dict()

    def _calc_ln_std(self, profile):
        # Depth based values
        ln_std = super()._calc_ln_std(profile)

        # Update based on soil_type name
        for name, value in self._ln_std_map.items():
            for i, l in enumerate(profile):
                if name in l.soil_type.name:
                    ln_std[i] = value

        return ln_std

    @classmethod
    def generic_model(cls, site_class, ln_std_map=None, **kwds):
        """Use generic model parameters based on site class.

        Parameters
        ----------
        site_class: str
            Site classification. Possible options are:
             * Geomatrix AB
             * Geomatrix CD
             * USGS AB
             * USGS CD
             * USGS A
             * USGS B
             * USGS C
             * USGS D

            See the report for definitions of the Geomatrix site
            classication. USGS site classification is based on :math:`V_{s30}`:

            =========== =====================
            Site Class  :math:`V_{s30}` (m/s)
            =========== =====================
            A           >750 m/s
            B           360 to 750 m/s
            C           180 to 360 m/s
            D           <180 m/s
            =========== =====================

        ln_std_map: dict[str, float], optional
            Mapping between the soil_type and the defined ln_std. Default is *None*.

        Returns
        -------
        DepthAndSoilTypeDependToroVelVariation
            Initialized :class:`DepthDependToroVelVariation` with generic parameters.
        """
        p = dict(cls.PARAMS[site_class])
        p.update(kwds)

        if "depth" not in kwds:
            p["depth"] = [0, 15]
            p["ln_std"] = [0.25, 0.15]

        p["ln_std_map"] = ln_std_map
        return cls(**p)


class SoilTypeVariation(object):
    def __init__(
        self,
        correlation,
        limits_mod_reduc=[0.05, 1],
        limits_damping=[0, 0.15],
        vary_bedrock=False,
    ):
        self._vary_bedrock = vary_bedrock
        self._correlation = correlation
        self._limits_mod_reduc = list(limits_mod_reduc)
        self._limits_damping = list(limits_damping)

    def __call__(self, soil_type):
        def get_values(nlp):
            try:
                return nlp.values
            except AttributeError:
                return np.asarray(nlp).astype(float)

        mod_reduc = get_values(soil_type.mod_reduc)
        damping = get_values(soil_type.damping)

        # A pair of correlated random variables
        randvar = randnorm.correlated(self.correlation)

        varied_mod_reduc, varied_damping = self._get_varied(randvar, mod_reduc, damping)

        # Clip the values to the specified min/max
        varied_mod_reduc = np.clip(
            varied_mod_reduc, self.limits_mod_reduc[0], self.limits_mod_reduc[1]
        )
        varied_damping = np.clip(
            varied_damping, self.limits_damping[0], self.limits_damping[1]
        )

        # Set the values
        realization = copy.deepcopy(soil_type)
        for attr_name, values in zip(
            ["mod_reduc", "damping"], [varied_mod_reduc, varied_damping]
        ):
            try:
                getattr(realization, attr_name).values = values
            except AttributeError:
                setattr(realization, attr_name, values)
        return realization

    def _get_varied(self, randvar, mod_reduc, damping):
        raise NotImplementedError

    @property
    def correlation(self):
        return self._correlation

    @property
    def limits_damping(self):
        return self._limits_damping

    @property
    def limits_mod_reduc(self):
        return self._limits_mod_reduc

    @property
    def vary_bedrock(self):
        return self._vary_bedrock


class DarendeliVariation(SoilTypeVariation):
    def _get_varied(self, randvar, mod_reduc, damping):
        mod_reduc_means = mod_reduc
        mod_reduc_stds = self.calc_std_mod_reduc(mod_reduc_means)
        varied_mod_reduc = mod_reduc_means + randvar[0] * mod_reduc_stds

        damping_means = damping
        damping_stds = self.calc_std_damping(damping_means)
        varied_damping = damping_means + randvar[1] * damping_stds

        return varied_mod_reduc, varied_damping

    @staticmethod
    def calc_std_mod_reduc(mod_reduc):
        """Calculate the standard deviation as a function of G/G_max.

        Equation 7.29 from Darendeli (2001).

        Parameters
        ----------
        mod_reduc : array_like
            Modulus reduction values.

        Returns
        -------
        std : :class:`numpy.ndarray`
            Standard deviation.
        """
        mod_reduc = np.asarray(mod_reduc).astype(float)
        std = np.exp(-4.23) + np.sqrt(
            0.25 / np.exp(3.62) - (mod_reduc - 0.5) ** 2 / np.exp(3.62)
        )
        return std

    @staticmethod
    def calc_std_damping(damping):
        """Calculate the standard deviation as a function of damping in decimal.

        Equation 7.30 from Darendeli (2001).

        Parameters
        ----------
        damping : array_like
            Material damping values in decimal.

        Returns
        -------
        std : :class:`numpy.ndarray`
            Standard deviation.
        """
        damping = np.asarray(damping).astype(float)
        std = (np.exp(-5) + np.exp(-0.25) * np.sqrt(100 * damping)) / 100.0
        return std


class SpidVariation(SoilTypeVariation):
    """Variation defined by the EPRI SPID (2013) and documented in
    PNNL (2014)."""

    def __init__(
        self,
        correlation,
        limits_mod_reduc=[0, 1],
        limits_damping=[0, 0.15],
        std_mod_reduc=0.15,
        std_damping=0.30,
    ):
        super().__init__(correlation, limits_mod_reduc, limits_damping)
        self._std_mod_reduc = std_mod_reduc
        self._std_damping = std_damping

    def _get_varied(self, randvar, mod_reduc, damping):
        # Vary the G/Gmax in transformed space.
        # Equation 9.43 of PNNL (2014)
        # Here epsilon is added so that the denomiator doesn't go to zero.
        f_mean = mod_reduc / (1 - mod_reduc + np.finfo(float).eps)
        # Instead of constraining the standard deviation at a specific
        # strain, then standard deviation is constrained at G/Gmax of 0.5.
        # This is modified from Equation 9.44 of PNNL (2014).
        f_std = self.std_mod_reduc * (1 / (1 - 0.5))
        f_real = np.exp(randvar[0] * f_std) * f_mean
        # Equation 9.45 of PNNL (2014)
        varied_mod_reduc = f_real / (1 + f_real)

        # Simple log distribution
        varied_damping = np.exp(randvar[1] * self.std_damping) * damping

        return varied_mod_reduc, varied_damping

    @property
    def std_damping(self):
        return self._std_damping

    @property
    def std_mod_reduc(self):
        return self._std_mod_reduc


def iter_varied_profiles(
    profile,
    count,
    var_thickness: ToroThicknessVariation = None,
    var_velocity: VelocityVariation = None,
    var_soiltypes: SoilTypeVariation = None,
):
    for i in range(count):
        # Copy the profile to form the realization
        p = profile

        if var_thickness:
            p = var_thickness(p)

        if var_velocity:
            p = var_velocity(p)

        if var_soiltypes:
            # Map of varied soil types
            varied = {str(st): var_soiltypes(st) for st in p.iter_soil_types()}
            # Create new layers
            end = None if var_soiltypes.vary_bedrock else -1
            layers = [
                site.Layer(varied[str(l.soil_type)], l.thickness, l.initial_shear_vel)
                for l in p[:end]
            ]
            # Add the unrandomized bedrock
            if not var_soiltypes.vary_bedrock:
                layers.append(p[-1])

            # Create a new profile
            p = site.Profile(layers, p.wt_depth)

        yield p



# The MIT License (MIT)
#
# Copyright (c) 2016-2021 Albert Kottke
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import collections

import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate
from matplotlib.colors import LogNorm
from matplotlib.colors import TwoSlopeNorm
from scipy.interpolate import interp1d

try:
    import pandas as pd
except ImportError:
    pd = None

import pykooh

from .motion import TimeSeriesMotion, WaveField, GRAVITY


def plot_amplification_evolv(
    calc,
    metric="accel_tf",
    depths=None,
    freqs=None,
    normalized=False,
    wave_field_out="within",
    diverging_cmap=True,
    include_vs_profile=False,
    ax=None,
    **kwds,
):
    # Default plotting kwds. Combine both set of plot keywords and prefer the provided
    kwds = {
        "cmap": "RdBu" if diverging_cmap else "magma_r",
        "shading": "gouraud",
        "norm": TwoSlopeNorm(vmin=0, vcenter=1) if diverging_cmap else LogNorm(),
    } | kwds

    if freqs is None:
        freqs = np.logspace(-1, 2, num=301)

    osc_damping = 0.05 if "osc_damping" not in kwds else kwds["osc_damping"]

    ln_freqs = np.log(freqs)
    ln_freqs_mot = np.log(calc.motion.freqs)

    def get_amp(metric, depth):
        loc_output = calc.profile.location(wave_field_out, depth=depth)
        if metric == "accel_tf":
            y = np.abs(calc.calc_accel_tf(calc.loc_input, loc_output))
            # Interpolate the specific frequencies
            y = np.interp(ln_freqs, ln_freqs_mot, y)
        elif metric == "site_amp":
            if get_amp.in_ars is None:
                get_amp.in_ars = calc.motion.calc_osc_accels(freqs, osc_damping)

            out_ars = calc.motion.calc_osc_accels(
                freqs, osc_damping, calc.calc_accel_tf(calc.loc_input, loc_output)
            )
            y = out_ars / get_amp.in_ars
        else:
            raise NotImplementedError

        return y

    # Initialize static variable
    get_amp.in_ars = None

    if depths is None:
        depths = np.linspace(0, calc.profile[-1].depth)

    if ax is None:
        fig, ax = plt.subplots()

    amps = np.array([get_amp(metric, d) for d in depths])
    if normalized:
        amps /= amps[-1, :]

    cf = ax.pcolormesh(freqs, depths, amps, **kwds)

    cb = plt.colorbar(cf, ax=ax)
    cb.set_label("|TF|" if metric == "accel_tf" else "Site Ampl.")

    ax.set(
        xlabel="Frequency (Hz)",
        xscale="log",
        ylabel="Depth (m)",
        yscale="linear",
        ylim=(depths[0], 0),
    )

    return ax


class OutputCollection(collections.abc.Collection):
    def __init__(self, outputs):
        super().__init__()
        self.outputs = outputs

    def __iter__(self):
        return iter(self.outputs)

    def __contains__(self, value):
        return value in self.outputs

    def __len__(self):
        return len(self.outputs)

    def __getitem__(self, key):
        return self.outputs[key]

    def __call__(self, calc, name=None):
        # Save results
        for o in self:
            o(calc, name=name)

    def reset(self):
        for o in self:
            o.reset()


def append_arrays(many, single):
    """Append an array to another padding with NaNs for constant length.

    Parameters
    ----------
    many : array_like of rank (j, k)
        values appended to a copy of this array. This may be a 1-D or 2-D
        array.
    single : array_like of rank l
        values to append. This should be a 1-D array.

    Returns
    -------
    append : :class:`numpy.ndarray`
        2-D array with rank (j + 1, max(k, l)) with missing values padded
        with :class:`numpy.nan`
    """
    assert np.ndim(single) == 1

    # Check if the values need to be padded to for equal length
    diff = single.shape[0] - many.shape[0]
    if diff < 0:
        single = np.pad(single, (0, -diff), "constant", constant_values=np.nan)
    elif diff > 0:
        # Need different padding based on if many is 1d or 2d.
        padding = ((0, diff), (0, 0)) if len(many.shape) > 1 else (0, diff)
        many = np.pad(many, padding, "constant", constant_values=np.nan)
    else:
        # No padding needed
        pass
    return np.c_[many, single]


class Output(object):
    _const_ref = False

    xscale = "log"
    yscale = "log"
    drawstyle = "default"

    def __init__(self, refs=None):
        self._refs = np.asarray([] if refs is None else refs)
        self._values = None
        self._names = []

    def __call__(self, calc, name=None):
        if name is None:
            if self.values is None:
                i = 1
            elif len(self.values.shape) == 1:
                i = 2
            else:
                i = self.values.shape[1] + 1
            name = "r%d" % i
        self._names.append(name)

    @property
    def refs(self):
        return self._refs

    @property
    def values(self):
        return self._values

    @property
    def names(self):
        return self._names

    def reset(self):
        self._values = None
        self._names = []
        if not self._const_ref:
            self._refs = np.array([])

    def iter_results(self):
        shared_ref = len(self.refs.shape) == 1
        for i, name in enumerate(self.names):
            refs = self.refs if shared_ref else self.refs[:, i]
            values = self.values if len(self.values.shape) == 1 else self.values[:, i]
            yield name, refs, values

    def _add_refs(self, refs):
        refs = np.asarray(refs)
        if len(self._refs) == 0:
            self._refs = np.array(refs)
        else:
            self._refs = append_arrays(self._refs, refs)

    def _add_values(self, values):
        values = np.asarray(values)
        if self._values is None:
            self._values = values
        else:
            self._values = append_arrays(self._values, values)

    def calc_stats(self, as_dataframe=False):
        ln_values = np.log(self.values)
        median = np.exp(np.nanmean(ln_values, axis=1))
        ln_std = np.nanstd(ln_values, axis=1)

        stats = {"ref": self.refs, "median": median, "ln_std": ln_std}
        if as_dataframe and pd:
            stats = pd.DataFrame(stats).set_index("ref")
            stats.index.name = self.ref_name

        return stats

    def to_dataframe(self):
        if not pd:
            raise RuntimeError("Install `pandas` library.")

        if isinstance(self.names[0], tuple):
            columns = pd.MultiIndex.from_tuples(self.names)
        else:
            columns = self.names

        df = pd.DataFrame(self.values, index=self.refs, columns=columns)

        return df

    @staticmethod
    def _get_xy(refs, values):
        return refs, values

    def plot(self, ax=None, style="indiv"):
        assert style in ["stats", "indiv"]

        if ax is None:
            fig, ax = plt.subplots()

        if style == "stats" and len(self.values.shape) > 1 and self.values.shape[1] < 3:
            raise RuntimeError("Unable to plot stats for less than 3 values.")

        if style == "stats":
            kwds = {"color": "C0", "alpha": 0.6, "lw": 0.8, "drawstyle": self.drawstyle}
        elif style == "indiv":
            kwds = {"lw": 1.0, "drawstyle": self.drawstyle}
        else:
            raise NotImplementedError("Valid options are: stats, indiv.")

        # Add the data
        x, y = self._get_xy(self.refs, self.values)
        lines = ax.plot(x, y, **kwds)

        if style == "stats":
            lines[0].set_label("Realization")
        else:
            for layer, name in zip(lines, self.names):
                layer.set_label(name)

        if style == "stats":
            stats = self.calc_stats()

            ax.plot(
                *self._get_xy(stats["ref"], stats["median"]),
                color="C1",
                lw=2,
                label="Median",
            )

        ax.set(
            xlabel=self.xlabel,
            xscale=self.xscale,
            ylabel=self.ylabel,
            yscale=self.yscale,
        )

        if len(lines) > 1:
            ax.legend()

        return ax


class OutputLocation(object):
    def __init__(self, wave_field, depth=None, index=None):
        self._depth = depth
        self._index = index
        if not isinstance(wave_field, WaveField):
            wave_field = WaveField[wave_field]
        self._wave_field = wave_field

    @property
    def depth(self):
        return self._depth

    @property
    def index(self):
        return self._index

    @property
    def wave_field(self):
        return self._wave_field

    def __call__(self, profile):
        """Lookup the location with the profile."""
        return profile.location(self.wave_field, depth=self.depth, index=self.index)


class LocationBasedOutput(Output):
    def __init__(self, ref, location):
        super().__init__(ref)
        self._location = location

    @property
    def location(self):
        return self._location

    def __call__(self, calc, name=None):
        raise NotImplementedError

    def _get_location(self, calc):
        """Locate location within the profile."""
        return self._location(calc.profile)


class TimeSeriesOutput(LocationBasedOutput):
    xlabel = "Time (sec)"
    xscale = "linear"
    ylabel = NotImplemented
    yscale = "linear"

    ref_name = "time"

    def __init__(self, location):
        super().__init__(None, location)

    @property
    def times(self):
        return self.refs

    def __call__(self, calc, name=None):
        if not isinstance(calc.motion, TimeSeriesMotion):
            raise NotImplementedError
        Output.__call__(self, calc, name)
        # Compute the response
        loc = self._get_location(calc)
        tf = self._get_trans_func(calc, loc)
        values = calc.motion.calc_time_series(tf)
        values = self._modify_values(calc, loc, values)
        self._add_values(values)
        # Add the reference
        refs = calc.motion.time_step * np.arange(len(values))
        self._add_refs(refs)

    def _get_trans_func(self, calc, location):
        raise NotImplementedError

    def _modify_values(self, calc, location, values):
        return values

    def to_dataframe(self):
        raise NotImplementedError


class AccelerationTSOutput(TimeSeriesOutput):
    ylabel = "Acceleration (g)"

    def _get_trans_func(self, calc, location):
        return calc.calc_accel_tf(calc.loc_input, location)


class AriasIntensityTSOutput(AccelerationTSOutput):
    ylabel = "Arias Intensity (m/s)"

    def _modify_values(self, calc, location, values):
        time_step = calc.motion.time_step
        values = scipy.integrate.cumtrapz(values ** 2, dx=time_step)
        values *= GRAVITY * np.pi / 2
        return values


class StrainTSOutput(TimeSeriesOutput):
    def __init__(self, location, in_percent=False):
        super().__init__(location)
        self._in_percent = in_percent
        assert self.location.wave_field == WaveField.within

    def _get_trans_func(self, calc, location):
        return calc.calc_strain_tf(calc.loc_input, location)

    def _modify_values(self, calc, location, values):
        if self._in_percent:
            # Convert to percent
            values *= 100.0
        return values

    @property
    def ylabel(self):
        suffix = "(%)" if self._in_percent else "(dec)"
        return "Shear Strain " + suffix


class StressTSOutput(TimeSeriesOutput):
    def __init__(self, location, damped=False, normalized=False):
        super().__init__(location)
        self._damped = damped
        self._normalized = normalized
        assert self.location.wave_field == WaveField.within

    @property
    def damped(self):
        return self._damped

    @property
    def ylabel(self):
        if self._normalized:
            ylabel = "Stress Ratio (τ/σ`ᵥ)"
        else:
            ylabel = "Stress (τ)"

        return ylabel

    def _get_trans_func(self, calc, location):
        tf = calc.calc_stress_tf(calc.loc_input, location, self.damped)

        if self._normalized:
            # Correct by effective stress at depth
            tf /= location.stress_vert(effective=True)

        return tf


class FourierAmplitudeSpectrumOutput(LocationBasedOutput):
    _const_ref = True
    xlabel = "Frequency (Hz)"
    ylabel = "Fourier Ampl. (cm/s)"

    ref_name = "freq"

    def __init__(self, freqs, location, ko_bandwidth=30):
        super().__init__(freqs, location)
        self._ko_bandwidth = ko_bandwidth

    @property
    def freqs(self):
        return self._refs

    @property
    def ko_bandwidth(self):
        return self._ko_bandwidth

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        loc = self._get_location(calc)
        tf = calc.calc_accel_tf(calc.loc_input, loc)

        if self.ko_bandwidth:
            smoothed = pykooh.smooth(
                self.freqs,
                calc.motion.freqs,
                np.abs(tf * calc.motion.fourier_amps),
                self.ko_bandwidth,
            )

        self._add_values(smoothed)


class ResponseSpectrumOutput(LocationBasedOutput):
    _const_ref = True
    xlabel = "Frequency (Hz)"

    ref_name = "freq"

    def __init__(self, freqs, location, osc_damping):
        super().__init__(freqs, location)
        self._osc_damping = osc_damping

    @property
    def freqs(self):
        return self._refs

    @property
    def periods(self):
        return 1.0 / np.asarray(self._refs)

    @property
    def osc_damping(self):
        return self._osc_damping

    @property
    def ylabel(self):
        return f"{100 * self.osc_damping:g}%-Damped, Spec. Accel. (g)"

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        loc = self._get_location(calc)
        tf = calc.calc_accel_tf(calc.loc_input, loc)
        ars = calc.motion.calc_osc_accels(self.freqs, self.osc_damping, tf)
        self._add_values(ars)


class RatioBasedOutput(Output):
    _const_ref = True

    def __init__(self, refs, location_in, location_out):
        super().__init__(refs)
        self._location_in = location_in
        self._location_out = location_out

    @property
    def location_in(self):
        return self._location_in

    @property
    def location_out(self):
        return self._location_out

    def __call__(self, calc, name=None):
        raise NotImplementedError

    def _get_locations(self, calc):
        """Locate locations within the profile."""
        return (self._location_in(calc.profile), self._location_out(calc.profile))


class AccelTransferFunctionOutput(RatioBasedOutput):
    xlabel = "Frequency (Hz)"
    ylabel = "Accel. Transfer Func."

    ref_name = "freq"

    def __init__(
        self, refs, location_in, location_out, ko_bandwidth=None, absolute=True
    ):
        super().__init__(refs, location_in, location_out)
        self._ko_bandwidth = ko_bandwidth
        self._absolute = absolute

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        # Locate position within the profile
        loc_in, loc_out = self._get_locations(calc)
        # Compute the response
        if self._absolute:
            tf = np.abs(calc.calc_accel_tf(loc_in, loc_out))
        else:
            tf = calc.calc_accel_tf(loc_in, loc_out)

        if self._ko_bandwidth is None:
            tf = np.interp(self.freqs, calc.motion.freqs, tf)
        else:
            tf = pykooh.smooth(self.freqs, calc.motion.freqs, tf, self._ko_bandwidth)

        self._add_values(tf)

    @property
    def freqs(self):
        return self._refs


class ResponseSpectrumRatioOutput(RatioBasedOutput):
    xlabel = "Frequency (Hz)"

    ref_name = "freq"

    def __init__(self, freqs, location_in, location_out, osc_damping):
        super().__init__(freqs, location_in, location_out)
        self._osc_damping = osc_damping

    @property
    def freqs(self):
        return self._refs

    @property
    def periods(self):
        return 1.0 / np.asarray(self._refs)

    @property
    def osc_damping(self):
        return self._osc_damping

    @property
    def ylabel(self):
        return f"{100 * self.osc_damping:g}%-Damped, Resp. Spectral Ratio"

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        loc_in, loc_out = self._get_locations(calc)
        in_ars = calc.motion.calc_osc_accels(
            self.freqs, self.osc_damping, calc.calc_accel_tf(calc.loc_input, loc_in)
        )
        out_ars = calc.motion.calc_osc_accels(
            self.freqs, self.osc_damping, calc.calc_accel_tf(calc.loc_input, loc_out)
        )
        ratio = out_ars / in_ars
        self._add_values(ratio)


class ProfileBasedOutput(Output):
    ylabel = "Depth (m)"
    yscale = "linear"
    drawstyle = "steps-post"

    ref_name = "depth"

    def __init__(self):
        super().__init__()

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        depths = np.r_[0, calc.profile.depth_mid[:-1]]
        self._add_refs(depths)

    def calc_stats(self, as_dataframe=False, ref=None):
        if ref is None:
            ref = np.linspace(0, np.nanmax(self.refs) * 1.05, num=512)

        n = self.values.shape[1]
        with np.errstate(divide="ignore", invalid="ignore"):
            # Ignore zeros in the data
            ln_values = np.array([self._ln_interp(i, ref) for i in range(n)]).T
            median = np.exp(np.nanmean(ln_values, axis=1))
            ln_std = np.nanstd(ln_values, axis=1)

        stats = {"ref": ref, "median": median, "ln_std": ln_std}
        if as_dataframe and pd:
            stats = pd.DataFrame(stats).set_index("ref")
            stats.index.name = self.ref_name

        return stats

    @staticmethod
    def _get_xy(refs, values):
        return values, refs

    def plot(self, ax=None, style="stats"):
        ax = Output.plot(self, ax, style)
        ax.invert_yaxis()
        return ax

    def _ln_interp(self, i, ref):
        """Interpolate the values in log-y space."""

        _ref = self.refs[:, i]
        # Only select points with valid entries
        mask = np.isfinite(_ref)
        _ref = _ref[mask]
        _ln_values = np.log(self.values[mask, i])

        if np.any(mask):
            f = interp1d(
                _ref,
                _ln_values,
                kind="next",
                fill_value=(_ln_values[0], _ln_values[-1]),
                bounds_error=False,
            )
            _ln_interped = f(ref)
        else:
            nans = np.empty_like(ref)
            nans[:] = np.nan
            _ln_interped = np.array(nans)

        return _ln_interped

    def to_dataframe(self, ref=None):
        if not pd:
            raise RuntimeError("Install `pandas` library.")

        if ref is None:
            ref = np.linspace(0, np.nanmax(self.refs))

        if isinstance(self.names[0], tuple):
            columns = pd.MultiIndex.from_tuples(self.names)
        else:
            columns = self.names

        # Ignore zeros in the data
        n = self.values.shape[1]
        values = np.exp(np.array([self._ln_interp(i, ref) for i in range(n)])).T

        df = pd.DataFrame(values, index=ref, columns=columns)

        return df


class MaxStrainProfile(ProfileBasedOutput):
    xlabel = "Max. Strain (dec)"

    def __init__(self):
        super().__init__()

    def __call__(self, calc, name=None):
        ProfileBasedOutput.__call__(self, calc, name)
        values = [0] + [layer.strain_max for layer in calc.profile[:-1]]
        self._add_values(values)


class DampingProfile(ProfileBasedOutput):
    xlabel = "Damping (dec)"

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        # Add depth at top of layer
        self._add_refs(calc.profile.depth)

        values = [layer.damping for layer in calc.profile[:-1]]
        # Bring the first mid-layer value to the surface
        values.insert(0, values[0])
        self._add_values(values)


class ShearModReducProfile(ProfileBasedOutput):
    xlabel = "G/Gmax"

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        # Add depth at top of layer
        self._add_refs(calc.profile.depth)

        values = [layer.shear_mod_reduc for layer in calc.profile[:-1]]
        # Bring the first mid-layer value to the surface
        values.insert(0, values[0])
        self._add_values(values)


class InitialVelProfile(ProfileBasedOutput):
    xlabel = "Initial Velocity (m/s)"

    def __init__(self):
        super().__init__()

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        # Add depth at top of layer
        self._add_refs(calc.profile.depth)

        values = [layer.initial_shear_vel for layer in calc.profile[:-1]]
        values.insert(0, values[0])
        self._add_values(values)


class CompatVelProfile(ProfileBasedOutput):
    xlabel = "Strain-Compatible Velocity (m/s)"

    def __init__(self):
        super().__init__()

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        # Add depth at top of layer
        self._add_refs(calc.profile.depth)

        values = [np.min(layer.shear_vel) for layer in calc.profile[:-1]]
        values.insert(0, values[0])
        self._add_values(values)


class CyclicStressRatioProfile(ProfileBasedOutput):
    # From Idriss and Boulanger (2008, pg. 70):
    # The 0.65 is a constant used to represent the reference stress
    # level. While being somewhat arbitrary it was selected in the
    # beginning of the development of liquefaction procedures in 1966
    # and has been in use ever since.
    _stress_level = 0.65

    def __init__(self):
        super().__init__()

    def __call__(self, calc, name=None):
        ProfileBasedOutput.__call__(self, calc, name)
        values = [
            layer.stress_shear_max / layer.stress_vert(layer.thickness / 2, True)
            for layer in calc.profile[:-1]
        ]
        # Repeat the first value for the surface
        values = self._stress_level * np.array([values[0]] + values)
        self._add_values(values)


class MaxAccelProfile(ProfileBasedOutput):
    xlabel = "Max. Accel. (g)"

    def __call__(self, calc, name=None):
        Output.__call__(self, calc, name)
        # Add depth at top of layer
        depths = calc.profile.depth
        values = [self._calc_accel(calc, depth) for depth in depths]
        self._add_refs(depths)
        self._add_values(values)

    def _calc_accel(self, calc, depth):
        return calc.motion.calc_peak(
            calc.calc_accel_tf(
                calc.loc_input, calc.profile.location("within", depth=depth)
            )
        )



#!/usr/bin/python3
# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup

with open("README.rst") as fp:
    readme = fp.read()

with open("HISTORY.rst") as fp:
    history = fp.read()

setup(
    name="pySRA",
    version="0.5.0",
    description="Site Response Analysis with Python",
    long_description=readme + "\n\n" + history,
    author="Albert Kottke",
    author_email="albert.kottke@gmail.com",
    url="http://github.com/arkottke/pysra",
    packages=find_packages(),
    install_requires=[
        "pystrata",
        "matplotlib",
        "numpy",
        "pyrvt",
        "pykooh",
        "scipy",
        "setuptools",
        "typing",
    ],
    extras_require={
        "dataframe": ["pandas"],
    },
    keywords="site response",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
    ],
    test_suite="tests",
)
